import strawberry
from typing import Generic, List, Optional, TypeVar, Union
from strawberry.field import StrawberryField
from strawberry.arguments import UNSET, is_unset, StrawberryArgument
from typing import List

from . import utils
from .arguments import argument
from .fields.types import is_auto

# for backward compatibility
from .legacy.filters import get_field_type, set_field_type

T= TypeVar("T")

@strawberry.input
class DjangoModelFilterInput:
    pk: strawberry.ID

@strawberry.input
class FilterLookup(Generic[T]):
    exact: Optional[T] = UNSET
    i_exact: Optional[T] = UNSET
    contains: Optional[T] = UNSET
    i_contains: Optional[T] = UNSET
    in_list: Optional[List[T]] = UNSET
    gt: Optional[T] = UNSET
    gte: Optional[T] = UNSET
    lt: Optional[T] = UNSET
    lte: Optional[T] = UNSET
    starts_with: Optional[T] = UNSET
    i_starts_with: Optional[T] = UNSET
    ends_with: Optional[T] = UNSET
    i_ends_with: Optional[T] = UNSET
    range: Optional[List[T]] = UNSET
    is_null: Optional[bool] = UNSET
    regex: Optional[str] = UNSET
    i_regex: Optional[str] = UNSET

lookup_name_conversion_map = {
    'i_exact': 'iexact',
    'i_contains': 'icontains',
    'in_list': 'in',
    'starts_with': 'startswith',
    'i_starts_with': 'istartswith',
    'ends_with': 'endswith',
    'i_ends_with': 'iendswith',
    'is_null': 'isnull',
    'i_regex': 'iregex',
}


def filter(model, *, name=None, lookups=False):
    try:
        import django_filters
    except ModuleNotFoundError:
        pass
    else:
        filterset_class = model
        if isinstance(filterset_class, django_filters.filterset.FilterSetMetaclass):
            utils.deprecated("support for 'django-filters' is deprecated and"
                " will removed in v0.3", stacklevel=2)
            from .legacy.filters import filter as filters_filter
            return filters_filter(filterset_class, name)

    def wrapper(cls):
        is_filter = lookups and 'lookups' or True
        from .type import process_type
        type_ = process_type(cls, model, is_input=True, partial=True, is_filter=is_filter)
        return type_
    return wrapper

def filter_deprecated(model, *, name=None, lookups=False):
    utils.deprecated("'strawberry_django.filter' is deprecated,"
        " use 'strawberry_django.filters.filter' instead", stacklevel=2)
    return filter(model, name=name, lookups=lookups)

def build_filter_kwargs(filters):
    filter_kwargs = {}
    filter_methods = []
    django_model = utils.get_django_model(filters)
    for field in utils.fields(filters):
        field_name = field.name
        field_value = getattr(filters, field_name)

        if is_unset(field_value):
            continue

        filter_method = getattr(filters, f'filter_{field_name}', None)
        if filter_method:
            filter_methods.append(filter_method)
            continue

        if django_model:
            if field_name not in django_model._meta._forward_fields_map:
                continue

        if field_name in lookup_name_conversion_map:
            field_name = lookup_name_conversion_map[field_name]
        if utils.is_strawberry_type(field_value):
            subfield_filter_kwargs, subfield_filter_methods = \
                build_filter_kwargs(field_value)
            for subfield_name, subfield_value in subfield_filter_kwargs.items():
                filter_kwargs[f'{field_name}__{subfield_name}'] = subfield_value
            filter_methods.extend(subfield_filter_methods)
        else:
            filter_kwargs[field_name] = field_value

    return filter_kwargs, filter_methods


def apply(filters, queryset, pk=UNSET):
    if not is_unset(pk):
        queryset = queryset.filter(pk=pk)
    if is_unset(filters) or filters is None:
        return queryset

    if hasattr(filters, 'filterset_class'):
        utils.deprecated("support for 'django-filters' is deprecated and"
            " will removed in v0.3", stacklevel=2)
        from .legacy.filters import apply as filters_apply
        return filters_apply(filters, queryset)

    filter_method = getattr(filters, 'filter', None)
    if filter_method:
        return filter_method(queryset)

    filter_kwargs, filter_methods = build_filter_kwargs(filters)
    queryset = queryset.filter(**filter_kwargs)
    for filter_method in filter_methods:
        queryset = filter_method(queryset=queryset)
    return queryset


class StrawberryDjangoFieldFilters:
    def __init__(self, filters=UNSET, **kwargs):
        self.filters = filters
        super().__init__(**kwargs)

    @property
    def arguments(self) -> List[StrawberryArgument]:
        arguments = []
        if not self.base_resolver:
            filters = self.get_filters()
            if self.django_model and not self.is_list:
                arguments.append(
                    argument('pk', strawberry.ID)
                )
            elif filters and not is_unset(filters):
                arguments.append(
                    argument('filters', filters)
                )
        return super().arguments + arguments

    def get_filters(self):
        if not is_unset(self.filters):
            return self.filters
        type_ = self.type or self.child.type
        if utils.is_django_type(type_):
            return type_._django_type.filters
        return None

    def get_queryset(self, queryset, info, pk=UNSET, filters=UNSET, **kwargs):
        queryset = apply(filters, queryset, pk)
        return super().get_queryset(queryset, info, **kwargs)



from ..registers import TypeRegister
from django.db.models.base import ModelBase

def resolve_type_args(args, types=None, is_input=False, single=False):
    arg_types = TypeRegister()
    models = []
    for arg in args:
        if isinstance(arg, ModelBase):
            models.append(arg)
            continue

        model = arg._django_model
        if model not in models:
            models.append(model)

        arg_types.register(model, arg)

    if not models:
        raise TypeError('No model defined for field generator')

    return_args = []
    for model in models:
        output_type = get_type(model, arg_types, types, is_input=False)
        if is_input:
            input_type = get_type(model, arg_types, types, is_input=True)
            return_args.append((model, output_type, input_type))
        else:
            return_args.append((model, output_type))
        if single:
            return return_args[0]
    return return_args

def get_type(model, arg_types, types, is_input):
    type = arg_types.get(model, is_input=is_input)
    if not type and types:
        type = types.get(model, is_input=is_input)
    if not type:
        raise TypeError(f"No type for model '{model._meta.object_name}'")
    return type



import strawberry
import typing
from .types import get_model_fields, update_fields
from .utils import deprecated
from ..fields.types import is_auto

_type = type

def type(model, *, fields=None, types=None, **kwargs):
    stacklevel = kwargs.get('is_input', False) and 4 or 3
    if 'fields' in kwargs:
        utils.deprecated("'fields' parameter is deprecated,"
            " please define all fields in class", stacklevel=stacklevel)
    if 'types' in kwargs:
        utils.deprecated("'types' parameter is deprecated,"
            " please define all types in class", stacklevel=stacklevel)
    def wrapper(cls):
        is_input = kwargs.get('is_input', False)
        partial = kwargs.pop('partial', False)
        model_fields = get_model_fields(cls, model, fields, types, is_input, partial)
        if not hasattr(cls, '__annotations__'):
            cls.__annotations__ = {}
        for field_name, field_type, field_value in model_fields:
            if field_name not in cls.__annotations__:
                cls.__annotations__[field_name] = field_type
            if not hasattr(cls, field_name):
                setattr(cls, field_name, field_value)
        for field_name, field_type in cls.__annotations__.items():
            if is_auto(field_type):
                raise TypeError(f"Field '{field_name}' has invalid type."
                    " Type 'auto' cannot be use together with"
                    " 'fields' parameter."
                )
        update_fields(cls, model)
        cls._django_model = model
        cls._partial = partial
        return strawberry.type(cls, **kwargs)
    return wrapper


def input(model, *, fields=None, types=None, partial=False, **kwargs):
    if 'is_update' in kwargs:
        deprecated("'is_update' argument is deprecated, please use 'partial' instead", stacklevel=2)
        partial = kwargs.pop('is_update')
    return type(model, fields=fields, types=types, partial=partial, is_input=True, **kwargs)


def generate_partial_input(model, input):
    if input._partial:
        return input
    cls = _type(f'{input.__name__}Update', (), { '__annotations__': {}})
    for field_name, field_type in input.__annotations__.items():
        field_value = getattr(input, field_name)
        if typing.get_origin(field_type) != typing.Optional:
            field_type = typing.Optional[field_type]
        cls.__annotations__[field_name] = field_type
        setattr(cls, field_name, field_value)
    cls._django_model = model
    return strawberry.type(cls, is_input=True)



import strawberry
from django.contrib import auth
from ..resolvers import django_resolver

@django_resolver
def resolve_login(info, username: str, password: str):
    request = info.context.request
    user = auth.authenticate(request, username=username, password=password)
    if user is not None:
        auth.login(request, user)
        return user
    auth.logout(request)
    return None

@django_resolver
def resolve_logout(info) -> bool:
    request = info.context.request
    ret = request.user.is_authenticated
    auth.logout(request)
    return ret

def login():
    mutation = strawberry.mutation(resolver=resolve_login)
    mutation.is_optional = True
    return mutation

def logout():
    return strawberry.mutation(resolver=resolve_logout)



import dataclasses
import django
import strawberry
from django.db import models
from strawberry.arguments import UNSET, is_unset
from typing import Any, Optional

from .fields.field import StrawberryDjangoField
from .fields.types import (
    is_optional,
    get_model_field, resolve_model_field_type, resolve_model_field_name,
)
from . import utils

_type = type

def get_type_attr(type_, field_name):
    attr = getattr(type_, field_name, UNSET)
    if utils.is_unset(attr):
        attr = getattr(type_, '__dataclass_fields__', {}).get(field_name, UNSET)
    return attr

def get_field(django_type, field_name, field_annotation=None):
    attr = get_type_attr(django_type.origin, field_name)

    if utils.is_field(attr):
        field = StrawberryDjangoField.from_field(attr, django_type)
    else:
        field = StrawberryDjangoField(
            default=attr,
            type_=field_annotation,
        )

    field.python_name = field_name
    if field_name in django_type.origin.__dict__.get('__annotations__', {}):
        # store origin django type for futher usage
        field.origin_django_type = django_type

    if field_annotation:
        # annotation of field is used as a class type
        field.type = field_annotation
        field.is_auto = utils.is_auto(field_annotation)

    try:
        # resolve the django_name and check if it is relation field. django_name
        # is used to access the field data in resolvers
        django_name = field.django_name or field_name
        model_field = get_model_field(django_type.model, django_name)
        field.django_name = resolve_model_field_name(model_field,
                django_type.is_input, django_type.is_filter)
        field.is_relation = model_field.is_relation
    except django.core.exceptions.FieldDoesNotExist:
        if field.django_name or field.is_auto:
            raise # field should exist, reraise catched exception
        model_field = None

    if field.is_relation:
        # change relation field type to auto if field is inherited from another
        # type. for example if field is inherited from output type but we are
        # configuring field for input type
        if not utils.is_similar_django_type(django_type, field.origin_django_type):
            field.is_auto = True

    if field.is_auto:
        # resolve type of auto field
        field.type = resolve_model_field_type(model_field, django_type)

    if is_optional(model_field, django_type.is_input, django_type.is_partial):
        field.type = Optional[field.type]
    
    if django_type.is_input:
        if field.default is dataclasses.MISSING:
            # strawberry converts UNSET value to MISSING, let's set
            # it back to UNSET. this is important especially for partial
            # input types
            #TODO: could strawberry support UNSET default value?
            field.default_value = UNSET
            field.default = UNSET

    return field


def get_fields(django_type):
    annotations = utils.get_annotations(django_type.origin)
    fields = {}

    # collect all annotated fields
    for field_name, field_annotation in annotations.items():
        field = get_field(django_type, field_name, field_annotation)
        fields[field_name] = field

    # collect non-annotated strawberry fields
    for field_name in dir(django_type.origin):
        if field_name in fields:
            continue
        attr = getattr(django_type.origin, field_name)
        if not utils.is_strawberry_field(attr):
            continue
        field = get_field(django_type, field_name)
        fields[field_name] = field

    return list(fields.values())



@dataclasses.dataclass
class StrawberryDjangoType:
    origin: Any
    model: Any
    is_input: bool
    is_partial: bool
    is_filter: bool
    filters: Any
    order: Any
    pagination: Any

def process_type(cls, model, *, filters=UNSET, pagination=UNSET, order=UNSET, **kwargs):
    original_annotations = cls.__dict__.get('__annotations__', {})

    django_type = StrawberryDjangoType(
        origin=cls,
        model=model,
        is_input=kwargs.get('is_input', False),
        is_partial=kwargs.pop('partial', False),
        is_filter=kwargs.pop('is_filter', False),
        filters=filters,
        order=order,
        pagination=pagination,
    )

    fields = get_fields(django_type)

    # update annotations and fields
    cls.__annotations__ = cls_annotations = {}
    for field in fields:
        cls_annotations[field.name] = field.type
        setattr(cls, field.name, field)

    strawberry.type(cls, **kwargs)

    # restore original annotations for further use
    cls.__annotations__ = original_annotations
    cls._django_type = django_type

    return cls
    

def type(model, *, filters=UNSET, **kwargs):
    if 'fields' in kwargs or 'types' in kwargs:
        from .legacy.type import type as type_legacy
        return type_legacy(model, **kwargs)

    def wrapper(cls):
        return process_type(cls, model, filters=filters, **kwargs)

    return wrapper


def input(model, *, partial=False, **kwargs):
    return type(model, partial=partial, is_input=True, **kwargs)

def mutation(model, **kwargs):
    return type(model, **kwargs)



from . import auth, filters, ordering
from .fields.field import field
from .fields.types import (
    auto, is_auto,
    DjangoFileType, DjangoImageType, DjangoModelType,
    OneToOneInput, OneToManyInput, ManyToOneInput, ManyToManyInput,
)
from .utils import fields
from .filters import filter_deprecated as filter
from .mutations.mutations import mutations
from .resolvers import django_resolver
from .type import type, input, mutation

# deprecated functions
from .legacy.mutations.auth import AuthMutation
from .legacy.queries.fields import queries
from .legacy.registers import TypeRegister



import strawberry
from . import resolvers
from .. import utils
from ..queries.arguments import resolve_type_args


def mutations(*args, types=None):
    type_args = resolve_type_args(args, types=types, is_input=True)
    mutation_fields = {}
    for model, output_type, input_type in type_args:
        object_name = utils.camel_to_snake(model._meta.object_name)
        mutation_fields[f'create_{object_name}'] = resolvers.create(model, output_type, input_type)
        mutation_fields[f'create_{object_name}s'] = resolvers.create_batch(model, output_type, input_type)
        mutation_fields[f'update_{object_name}s'] = resolvers.update(model, output_type, input_type)
        mutation_fields[f'delete_{object_name}s'] = resolvers.delete(model, output_type, input_type)
    return strawberry.type(type('Mutation', (), mutation_fields))

mutations.create = resolvers.create
mutations.create_batch = resolvers.create_batch
mutations.update = resolvers.update
mutations.delete = resolvers.delete 



from typing import List, Optional
import strawberry
from . import resolvers
from .. import utils
from .arguments import resolve_type_args

def queries(*args, types=None):
    type_args = resolve_type_args(args, types=types)
    query_fields = {}
    for model, object_type in type_args:
        object_name = utils.camel_to_snake(model._meta.object_name)
        query_fields[f'{object_name}'] = resolvers.get_object_resolver(model, object_type)
        query_fields[f'{object_name}s'] = resolvers.get_list_resolver(model, object_type)
    return strawberry.type(type('Query', (), query_fields))

queries.get = resolvers.get_object_resolver
queries.list = resolvers.get_list_resolver



class TypeRegister:
    def __init__(self):
        self.generic = {}
        self.types = {}
        self.inputs = {}

    # key can be field name, django model field or django model
    def register(self, key, type=None):
        django_model = getattr(key, '_django_model', None)
        if type:
            self.add(key, type)
            return type

        if django_model:
            key, type = django_model, key
            self.add(key, type)
            return type

        def wrapper(type):
            self.add(key, type)
            return type
        return wrapper

    def add(self, key, type):
        if hasattr(type, '_type_definition'):
            if type._type_definition.is_input:
                self.inputs[key] = type
            else:
                self.types[key] = type
        else:
            self.generic[key] = type

    def get(self, key, is_input, default=None):
        if is_input:
            type = self.inputs.get(key, default)
        else:
            type = self.types.get(key, default)
        if type is default:
            type = self.generic.get(key, default)
        return type



import enum
import strawberry
import strawberry_django
from strawberry.arguments import UNSET, is_unset, StrawberryArgument
from typing import Optional, List

from .arguments import argument
from . import utils


@strawberry.enum
class Ordering(enum.Enum):
    ASC = 'ASC'
    DESC = 'DESC'


def generate_order_args(order, prefix=''):
    args = []
    for field in strawberry_django.fields(order):
        ordering = getattr(order, field.name, utils.UNSET)
        if utils.is_unset(ordering):
            continue
        if ordering == Ordering.ASC:
            args.append(f'{prefix}{field.name}')
        elif ordering == Ordering.DESC:
            args.append(f'-{prefix}{field.name}')
        else:
            prefix = f'{prefix}{field.name}__'
            subargs = generate_order_args(ordering, prefix=prefix)
            args.extend(subargs)
    return args

def order(model):
    def wrapper(cls):
        for name, type_ in cls.__annotations__.items():
            if strawberry_django.is_auto(type_):
                type_ = Ordering
            cls.__annotations__[name] = Optional[type_]
            setattr(cls, name, utils.UNSET)
        return strawberry.input(cls)
    return wrapper


def apply(order, queryset):
    if utils.is_unset(order) or order is None:
        return queryset
    args = generate_order_args(order)
    if not args:
        return queryset
    return queryset.order_by(*args)


class StrawberryDjangoFieldOrdering:
    def __init__(self, order=None, **kwargs):
        self.order = order
        super().__init__(**kwargs)

    @property
    def arguments(self) -> List[StrawberryArgument]:
        arguments = []
        if not self.base_resolver:
            if self.order:
                arguments.append(
                    argument('order', self.order)
                )
        return super().arguments + arguments

    def get_queryset(self, queryset, info, order=UNSET, **kwargs):
        queryset = apply(order, queryset)
        return super().get_queryset(queryset, info, **kwargs)



from typing import List, Optional
import strawberry
from .. import fields, hooks, utils
from ..type import generate_partial_input
from ..queries.arguments import resolve_type_args
from ...resolvers import django_resolver

def create(*args, types=None, pre_save=None, post_save=None):
    model, output_type, input_type = resolve_type_args(args, types=types, is_input=True, single=True)
    @hooks.add(pre_save=pre_save, post_save=post_save)
    @strawberry.mutation
    @django_resolver
    def mutation(info, data: input_type) -> output_type:
        instance_data = utils.get_input_data(model, data)
        instance = model(**instance_data)
        def caller(hook):
            hook(info, instance)
        mutation._call_hooks('pre_save', caller)
        instance.save()
        update_m2m_fields(model, [instance], data)
        mutation._call_hooks('post_save', caller)
        return instance
    return mutation

def create_batch(*args, types=None, pre_save=None, post_save=None):
    model, output_type, input_type = resolve_type_args(args, types=types, is_input=True, single=True)
    @hooks.add(pre_save=pre_save, post_save=post_save)
    @strawberry.mutation
    @django_resolver
    def mutation(data: List[input_type]) -> List[output_type]:
        instances = []
        for d in data:
            instance_data = utils.get_input_data(model, d)
            instance = model(**instance_data)
            def caller(hook):
                hook(info=info, instance=instance)
            mutation._call_hooks('pre_save', caller)
            instance.save()
            update_m2m_fields(model, [instance], data)
            mutation._call_hooks('post_save', caller)
            instances.append(instance)
        return instances
    return mutation

def update(*args, types=None):
    model, output_type, input_type = resolve_type_args(args, types=types, is_input=True, single=True)
    update_type = generate_partial_input(model, input_type)
    @strawberry.mutation
    @django_resolver
    def mutation(data: update_type, filters: Optional[List[str]] = []) -> List[output_type]:
        qs = model.objects.all()
        if filters:
            filter, exclude = utils.process_filters(filters)
            qs = qs.filter(**filter).exclude(**exclude)
        update_data = utils.get_input_data(model, data)
        qs.update(**update_data)
        update_m2m_fields(model, qs, data)
        return qs.all()
    return mutation

def delete(*args, types=None):
    model, output_type, input_type = resolve_type_args(args, types=types, is_input=True, single=True)
    @strawberry.mutation
    @django_resolver
    def mutation(filters: Optional[List[str]] = []) -> List[strawberry.ID]:
        qs = model.objects.all()
        if filters:
            filter, exclude = utils.process_filters(filters)
            qs = qs.filter(**filter).exclude(**exclude)
        ids = list(qs.values_list('id', flat=True))
        qs.delete()
        return ids
    return mutation


#internal helpers

def update_m2m_fields(model, objects, data):
    data = utils.get_input_data_m2m(model, data)
    if not data:
        return
    # iterate through objects and update m2m fields
    for obj in objects:
        for key, actions in data.items():
            relation_field = getattr(obj, key)
            for key, values in actions.items():
                if key == 'add':
                    relation_field.add(*values)
                elif key == 'set':
                    relation_field.set(values)
                elif key == 'remove':
                    relation_field.remove(*values)



from strawberry.arguments import StrawberryArgument, UNSET, convert_arguments, is_unset
from typing import List
from asgiref.sync import sync_to_async
from ..resolvers import django_resolver
from .. import utils, types
from ..fields.field import StrawberryDjangoFieldBase, StrawberryDjangoFieldFilters, StrawberryField
from ..fields.types import OneToOneInput, OneToManyInput, ManyToOneInput, ManyToManyInput

from django.db import transaction

class DjangoMutationBase:
    def __init__(self, input_type, **kwargs):
        self.input_type = input_type
        self.many = False
        super().__init__(graphql_name=None, python_name=None, type_=None, **kwargs)

    def post_init(self):
        type_ = self.type or self.child.type
        if self.input_type:
            assert self.django_model == utils.get_django_model(self.input_type), (
                'Input and output types should be from the same Django model')

    @property
    def arguments(self):
        arguments = []
        if self.input_type:
            arguments.append(get_argument('data', self.input_type, self.many))
        return arguments + super().arguments

    @django_resolver
    def get_result(self, source, info, kwargs):
        kwargs = convert_arguments(kwargs, self.arguments)
        return self.resolver(info=info, source=source, **kwargs)

    def get_wrapped_resolver(self):
        self.post_init()
        return super().get_wrapped_resolver()


class DjangoCreateMutation(
        DjangoMutationBase,
        StrawberryDjangoFieldBase,
        StrawberryField):

    def post_init(self):
        if self.is_list:
            self.many = True
        super().post_init()

    def create(self, data):
        input_data = get_input_data(self.input_type, data)
        instance = self.django_model.objects.create(**input_data)
        update_m2m([instance], data)
        return instance

    @transaction.atomic
    def resolver(self, info, source, data):
        if self.many:
            return [self.create(d) for d in data]
        else:
            return self.create(data)


class DjangoUpdateMutation(
        DjangoMutationBase,
        StrawberryDjangoFieldFilters,
        StrawberryDjangoFieldBase,
        StrawberryField):

    @transaction.atomic
    def resolver(self, info, source, data, **kwargs):
        queryset = self.django_model.objects.all()
        queryset = self.get_queryset(queryset=queryset, info=info, data=data, **kwargs)
        input_data = get_input_data(self.input_type, data)
        queryset.update(**input_data)
        update_m2m(queryset, data)
        return queryset


class DjangoDeleteMutation(
        DjangoMutationBase,
        StrawberryDjangoFieldFilters,
        StrawberryDjangoFieldBase,
        StrawberryField):

    @transaction.atomic
    def resolver(self, info, source, **kwargs):
        queryset = self.django_model.objects.all()
        queryset = self.get_queryset(queryset=queryset, info=info, **kwargs)
        instances = list(queryset)
        queryset.delete()
        return instances


def get_argument(name, type_, is_list=False):
    if is_list:
        return StrawberryArgument(
            python_name = name,
            graphql_name = name,
            type_ = None,
            child = StrawberryArgument(graphql_name=None, python_name=None, type_=type_),
            is_list = True,
        )
    else:
        return StrawberryArgument(
            python_name = name,
            graphql_name = name,
            type_ = type_,
        )


def get_input_data(input_type, data):
    input_data = {}
    m2m_data = {}
    for field in input_type._type_definition.fields:
        value = getattr(data, field.name)
        if isinstance(value, (ManyToOneInput, ManyToManyInput)):
            continue
        if is_unset(value):
            continue
        if isinstance(value, OneToManyInput):
            value = value.set
        input_data[field.django_name] = value
    return input_data


def update_m2m(queryset, data):
    #TODO: optimize
    for field_name, field_value in vars(data).items():
        if not isinstance(field_value, (ManyToOneInput, ManyToManyInput)):
            continue

        for instance in queryset:
            f = getattr(instance, field_name)
            if not is_unset(field_value.set):
                if field_value.add:
                    raise ValueError("'add' cannot be used together with 'set'")
                if field_value.remove:
                    raise ValueError("'remove' cannot be used together with 'set'")

                values = field_value.set
                if isinstance(field_value, ManyToOneInput):
                    values = [f.model.objects.get(pk=pk) for pk in values]
                if values:
                    f.set(values)
                else:
                    f.clear()
            else:
                if field_value.add:
                    values = field_value.add
                    if isinstance(field_value, ManyToOneInput):
                        values = [f.model.objects.get(pk=pk) for pk in values]
                    f.add(*values)
                if field_value.remove:
                    values = field_value.remove
                    if isinstance(field_value, ManyToOneInput):
                        values = [f.model.objects.get(pk=pk) for pk in values]
                    f.remove(*values)



import strawberry_django

def resolve_current_user(info):
    if not info.context.request.user.is_authenticated:
        return None
    return info.context.request.user

def current_user():
    field = strawberry_django.field(resolver=resolve_current_user)
    field.is_optional = True
    return field



from django.db import models
from typing import List, Optional
import inspect
import strawberry
from .. import hooks, utils
from ...resolvers import django_resolver
from .arguments import resolve_type_args


def get_object_resolver(*args, types=None):
    model, object_type = resolve_type_args(args, types=types, single=True)
    @strawberry.field
    @django_resolver
    def resolver(id: strawberry.ID) -> object_type:
        obj = model.objects.get(id=id)
        return obj
    return resolver


def get_list_resolver(*args, types=None, queryset=None):
    model, object_type = resolve_type_args(args, types=types, single=True)
    @hooks.add(queryset=queryset)
    @strawberry.field
    @django_resolver
    def resolver(info, filters: Optional[List[str]] = [], order_by: Optional[List[str]] = []) -> List[object_type]:
        class context:
            qs = model.objects.all()
        if filters:
            filter, exclude = utils.process_filters(filters)
            context.qs = context.qs.filter(**filter).exclude(**exclude)
        if order_by:
            context.qs = context.qs.order_by(*order_by)
        def queryset(hook):
            context.qs = hook(info=info, qs=context.qs)
        resolver._call_hooks('queryset', queryset)
        return context.qs
    return resolver


def get_resolver(resolver=None, field_name=None, is_relation=False, is_m2m=False):
    if resolver:
        if inspect.iscoroutinefunction(resolver):
            return resolver
        return django_resolver(resolver)

    if not is_relation:
        if field_name:
            def resolver(root):
                return getattr(root, field_name)
            return resolver
        return None

    if is_m2m:
        def resolver(root, info, filters: Optional[List[str]] = [], order_by: Optional[List[str]] = []):
            return get_instance_field(root, field_name, info, filters, order_by)
        return resolver

    else:
        def resolver(root, info):
            return get_instance_field(root, field_name, info)
        return resolver

@django_resolver
def get_instance_field(instance, field_name, info=None, filters=None, order_by=None):
    attr = getattr(instance, field_name or info.field_name)
    if not isinstance(attr, (models.QuerySet, models.Manager)):
        return attr

    # m2m
    qs = attr.all()
    if filters:
        filter, exclude = utils.process_filters(filters)
        qs = qs.filter(**filter).exclude(**exclude)
    if order_by:
        qs = qs.order_by(*order_by)
    return qs




from strawberry.arguments import UNSET
from ..legacy.mutations.fields import mutations as mutations_legacy
from .fields import DjangoCreateMutation, DjangoUpdateMutation, DjangoDeleteMutation

def mutations(*args, **kwargs):
    return mutations_legacy(*args, **kwargs)

def create(input_type=UNSET, *args, types=None, pre_save=None, post_save=None):
    if args or types:
        args = (input_type,) + args
        return mutations_legacy.create(*args, types=types, pre_save=pre_save, post_save=post_save)
    return DjangoCreateMutation(input_type)

def update(input_type=UNSET, *args, filters=UNSET, types=None):
    if args or types:
        args = (input_type,) + args
        return mutations_legacy.update(*args, types=types)
    return DjangoUpdateMutation(input_type, filters=filters)

def delete(*args, types=None, filters=UNSET):
    if args or types:
        return mutations_legacy.delete(*args, types=types)
    return DjangoDeleteMutation(input_type=None, filters=filters)

mutations.create = create
mutations.update = update
mutations.delete = delete



from django.db import models
from asgiref.sync import sync_to_async
import functools
import inspect
from . import utils


def django_resolver(resolver):
    """This decorator is used to make sure that resolver is always called from
    sync context.  sync_to_async helper in used if function is called from
    async context. This is useful especially with Django ORM, which does not
    support async. Coroutines are not wrapped.
    """

    if inspect.iscoroutinefunction(resolver) or inspect.isasyncgenfunction(resolver):
        return resolver

    @functools.wraps(resolver)
    def wrapper(*args, **kwargs):
        if utils.is_async():
            return call_sync_resolver(resolver, *args, **kwargs)
        else:
            return resolver(*args, **kwargs)

    return wrapper


def sync_to_async_thread_sensitive(func):
    # django 3.0 defaults to thread_sensitive=False
    return sync_to_async(func, thread_sensitive=True)


@sync_to_async_thread_sensitive
def call_sync_resolver(resolver, *args, **kwargs):
    """This function executes resolver function in sync context and ensures
    that querysets are executed.
    """

    result = resolver(*args, **kwargs)
    if isinstance(result, models.QuerySet):
        result = list(result)
    return result



import functools
from typing import Type, Optional, List
import strawberry
from graphql import GraphQLError
from strawberry.arguments import UNSET, is_unset


__all__ = [
    "InvalidFilterError",
    "apply",
    "filter",
    "get_field_type",
    "set_field_type",
]


class DummyDjangoFilters:
    def __getattribute__(self, attr):
        # make mocker happy
        if attr == '__enter__':
            raise AttributeError

try:
    import django_filters
except ModuleNotFoundError:
    django_filters = DummyDjangoFilters()


def assert_django_filters_installed(fn):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        if isinstance(django_filters, DummyDjangoFilters):
            raise ModuleNotFoundError(
                'You need to install django-filter to use "strawberry_django.filter". '
                "See https://django-filter.readthedocs.io/"
            )
        return fn(*args, **kwargs)
    return wrapper


# TODO
#  Some projects might want to change this to eg UUID
#  We should make it configurable
#  In the meantime they can overwrite "type_map" using "set_filter_field_type"
ID_TYPE = strawberry.ID


filter_field_type_map = {
    django_filters.BooleanFilter: Optional[bool],
    django_filters.CharFilter: Optional[str],
    django_filters.ChoiceFilter: Optional[str],
    django_filters.DateFilter: Optional[str],
    django_filters.DateTimeFilter: Optional[str],
    django_filters.DurationFilter: Optional[str],
    django_filters.IsoDateTimeFilter: Optional[str],
    django_filters.ModelChoiceFilter: Optional[ID_TYPE],
    django_filters.ModelMultipleChoiceFilter: Optional[List[ID_TYPE]],
    django_filters.MultipleChoiceFilter: Optional[List[str]],
    # Use str for number fields, because it might be int, decimal or float, and casting
    # it to any of those types might cause incorrect results for the other types.
    django_filters.NumberFilter: Optional[str],
    django_filters.TimeFilter: Optional[str],
    django_filters.UUIDFilter: Optional[str],

    # Not implemented because its difficult or impossible to create input types:
    # django_filters.AllValuesFilter
    # django_filters.AllValuesMultipleFilter
    # django_filters.DateFromToRangeFilter
    # django_filters.DateRangeFilter
    # django_filters.DateTimeFromToRangeFilter
    # django_filters.IsoDateTimeFromToRangeFilter
    # django_filters.OrderingFilter
    # django_filters.RangeFilter
    # django_filters.NumericRangeFilter
    # django_filters.TimeRangeFilter
    # django_filters.LookupChoiceFilter
    # django_filters.TypedChoiceFilter
    # django_filters.TypedMultipleChoiceFilter
}


class InvalidFilterError(GraphQLError):
    pass


@assert_django_filters_installed
def get_field_type(field_type):
    if type(field_type) != type:
        raise TypeError(
            f"expected 'type', received {type(field_type)}. Maybe you forgot to call type()?"
        )
    try:
        return filter_field_type_map[field_type]
    except KeyError:
        raise TypeError(f"No type defined for field type '{field_type}'")


@assert_django_filters_installed
def set_field_type(field_type, to_type):
    if type(field_type) != type:
        raise TypeError(
            f"expected 'type', received {type(field_type)}. Maybe you forgot to call type()?"
        )
    filter_field_type_map[field_type] = to_type


@assert_django_filters_installed
def apply(filter_instance, queryset):
    if is_unset(filter_instance) or not filter_instance:
        return queryset

    data = {}
    for field_name in filter_instance.filterset_class.get_fields():
        value = getattr(filter_instance, field_name, None)
        if not is_unset(value):
            data[field_name] = value

    filterset = filter_instance.filterset_class(
        data=data,
        queryset=queryset,
    )

    if not filterset.is_valid():
        raise InvalidFilterError(filterset.errors)

    return filterset.qs


@assert_django_filters_installed
def filter(filterset_class: Type[django_filters.FilterSet], name=None):
    if not isinstance(filterset_class, django_filters.filterset.FilterSetMetaclass):
        raise TypeError(
            "strawberry_django.filter expects a class that inherits django_filters.FilterSet, received %s",
            type(filterset_class),
        )

    filters = filterset_class.get_filters()
    name = name or filterset_class.__name__
    cls = type(name, (), {"__annotations__": {}, "filterset_class": filterset_class})

    for field_name in filterset_class.get_filters().keys():
        filter_field = filters[field_name]
        field_type = get_field_type(type(filter_field))
        cls.__annotations__[field_name] = field_type
        setattr(cls, field_name, UNSET)

    return strawberry.input(cls)



import dataclasses
import strawberry
from strawberry.arguments import UNSET, is_unset, StrawberryArgument
from typing import List

from . import utils
from .arguments import argument

@strawberry.input
class OffsetPaginationInput:
    offset: int = 0
    limit: int = -1

def apply(pagination, queryset):
    if is_unset(pagination) or pagination is None:
        return queryset
    start = pagination.offset
    stop = start + pagination.limit
    return queryset[start:stop]

class StrawberryDjangoPagination:
    def __init__(self, pagination=UNSET, **kwargs):
        self.pagination = pagination
        super().__init__(**kwargs)

    @property
    def arguments(self) -> List[StrawberryArgument]:
        arguments = []
        if not self.base_resolver:
            pagination = self.get_pagination()
            if pagination and not is_unset(pagination):
                arguments.append(
                    argument('pagination', OffsetPaginationInput)
                )
        return super().arguments + arguments

    def get_pagination(self):
        if not is_unset(self.pagination):
            return self.pagination
        type_ = self.type or self.child.type
        if utils.is_django_type(type_):
            return type_._django_type.pagination
        return None

    def get_queryset(self, queryset, info, pagination=UNSET, **kwargs):
        queryset = apply(pagination, queryset)
        return super().get_queryset(queryset, info, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['strawberry_django',
 'strawberry_django.auth',
 'strawberry_django.fields',
 'strawberry_django.legacy',
 'strawberry_django.legacy.mutations',
 'strawberry_django.legacy.queries',
 'strawberry_django.mutations']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3.0', 'strawberry-graphql>=0.64.4']

setup_kwargs = {
    'name': 'strawberry-graphql-django',
    'version': '0.2.2',
    'description': 'Strawberry GraphQL Django extension',
    'long_description': "# Strawberry GraphQL Django extension\n\n[![CI](https://github.com/la4de/strawberry-graphql-django/actions/workflows/main.yml/badge.svg)](https://github.com/la4de/strawberry-graphql-django/actions/workflows/main.yml)\n[![PyPI](https://img.shields.io/pypi/v/strawberry-graphql-django)](https://pypi.org/project/strawberry-graphql-django/)\n[![PyPI - Downloads](https://img.shields.io/pypi/dm/strawberry-graphql-django)](https://pypi.org/project/strawberry-graphql-django/)\n\nThis package provides simple and powerful tools to generate GraphQL types, queries, mutations and resolvers from Django models.\n\nInstalling `strawberry-graphql-django` package from the python package repository.\n```shell\npip install strawberry-graphql-django\n```\n\nFull documentation is available under [docs](https://github.com/strawberry-graphql/strawberry-graphql-django/tree/main/docs/index.md) github folder.\n\n\n## Supported features\n\n* GraphQL type generation from models\n* Filtering, pagination and ordering\n* Basic create, retrieve, update and delete (CRUD) types and mutations\n* Basic Django auth support, current user query, login and logout mutations\n* Django sync and async views\n* Unit test integration\n\n\n## Basic Usage\n\n```python\n# models.py\nfrom django.db import models\n\nclass Fruit(models.Model):\n    name = models.CharField(max_length=20)\n    color = models.ForeignKey('Color', blank=True, null=True,\n            related_name='fruits', on_delete=models.CASCADE)\n\nclass Color(models.Model):\n    name = models.CharField(max_length=20)\n```\n\n```python\n# types.py\nimport strawberry\nfrom strawberry.django import auto\nfrom typing import List\nfrom . import models\n\n@strawberry.django.type(models.Fruit)\nclass Fruit:\n    id: auto\n    name: auto\n    color: 'Color'\n\n@strawberry.django.type(models.Color)\nclass Color:\n    id: auto\n    name: auto\n    fruits: List[Fruit]\n```\n\n```python\n# schema.py\nimport strawberry\nfrom typing import List\nfrom .types import Fruit\n\n@strawberry.type\nclass Query:\n    fruits: List[Fruit] = strawberry.django.field()\n\nschema = strawberry.Schema(query=Query)\n```\n\nCode above generates following schema.\n\n```schema\ntype Fruit {\n  id: ID!\n  name: String!\n  color: Color\n}\n\ntype Color {\n  id: ID!\n  name: String!\n  fruits: [Fruit!]\n}\n\ntype Query {\n  fruits: [Fruit!]!\n}\n```\n\n```python\n# urls.py\nfrom django.urls import include, path\nfrom strawberry.django.views import AsyncGraphQLView\nfrom .schema import schema\n\nurlpatterns = [\n    path('graphql', AsyncGraphQLView.as_view(schema=schema)),\n]\n```\n\nSee complete Django project from github repository folder [examples/django](https://github.com/strawberry-graphql/strawberry-graphql-django/tree/main/examples/django).\n\n\n## Running unit tests\n```\npoetry install\npoetry run pytest\n```\n\n## Contributing\n\nWe are happy to get pull requests and feedback from you.\n",
    'author': 'Lauri Hintsala',
    'author_email': 'lauri.hintsala@verkkopaja.fi',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/strawberry-graphql/strawberry-graphql-django',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import dataclasses
import strawberry
from strawberry.arguments import is_unset, UNSET
from strawberry.field import StrawberryField
from django.db import models
import asyncio
import warnings


def is_async():
    # django uses the same method to detect async operation
    # https://github.com/django/django/blob/76c0b32f826469320c59709d31e2f2126dd7c505/django/utils/asyncio.py
    try:
        event_loop = asyncio.get_event_loop()
    except RuntimeError:
        pass
    else:
        if event_loop.is_running():
            return True
    return False


def deprecated(msg, stacklevel=1):
    warnings.warn(msg, DeprecationWarning, stacklevel=stacklevel + 1)

def is_strawberry_type(obj):
    return hasattr(obj, '_type_definition')

def is_strawberry_field(obj):
    return isinstance(obj, StrawberryField)

def is_strawberry_django_field(obj):
    from strawberry_django.fields.field import StrawberryDjangoField
    return isinstance(obj, StrawberryDjangoField)

def is_django_type(obj):
    return hasattr(obj, '_django_type')

def is_django_model(obj):
    return isinstance(obj, models.base.ModelBase)

def is_field(obj):
    return isinstance(obj, dataclasses.Field)

def is_django_field(obj):
    from .fields.field import DjangoField
    return isinstance(obj, DjangoField)

def fields(obj):
    return obj._type_definition.fields

def is_auto(obj):
    from .fields.types import is_auto
    return is_auto(obj)

def get_django_model(type_):
    if not is_django_type(type_):
        return
    return type_._django_type.model

def is_similar_django_type(a, b):
    if not a or not b:
        return False
    if a.is_input != b.is_input:
        return False
    if a.is_filter != b.is_filter:
        return False
    return True

def get_annotations(cls):
    annotations = {}
    for c in reversed(cls.__mro__):
        if '__annotations__' in c.__dict__:
            annotations.update(c.__annotations__)
    return annotations



from strawberry.arguments import is_unset, UNSET
from ..utils import *

import ast

def parse_value(value):
    try:
        return ast.literal_eval(value)
    except ValueError:
        raise ValueError('Invalid filter value')

def process_filters(filters):
    filter, exclude = {}, {}
    for string in filters:
        try:
            k, v = string.split('=', 1)
        except ValueError:
            raise ValueError(f'Invalid filter "{filter}"')
        if '!' in k:
            k = k.strip('!')
            exclude[k] = parse_value(v)
        else:
            filter[k] = parse_value(v)
    return filter, exclude

def get_input_data(model, data):
    values = {}
    for field in model._meta.fields:
        field_name = field.attname
        value = getattr(data, field_name, UNSET)
        if is_unset(value):
            continue
        values[field_name] = value
    return values

def get_input_data_m2m(model, data):
    values = {}
    for field in model._meta.many_to_many:
        for action in ('add', 'set', 'remove'):
            field_name = field.attname
            value = getattr(data, f'{field_name}_{action}', UNSET)
            if is_unset(value):
                continue
            actions = values.setdefault(field_name, {})
            actions[action] = value
    return values

def camel_to_snake(s):
    return ''.join(['_'+c.lower() if c.isupper() else c for c in s]).lstrip('_')

def snake_to_camel(s):
    return s.title().replace('_', '')



from strawberry.arguments import StrawberryArgument, UNSET

def argument(name, type_, is_optional=True, default=UNSET):
    return StrawberryArgument(
        default=default,
        description=None,
        graphql_name=None,
        is_optional=is_optional,
        origin=None,
        python_name=name,
        type_=type_,
    )




from strawberry.arguments import is_unset, UNSET
from typing import List, Optional
import strawberry
from . import utils
from .fields import DjangoField, field as strawberry_django_field

from ..fields.types import field_type_map, is_optional

class LazyModelType(strawberry.LazyType):
    def __init__(self, field, type_register, is_input):
        self.field, self.type_register, self.is_input = field, type_register, is_input
        super().__init__(field.name, None, None)

    def resolve_type(self):
        model = self.field.related_model

        field_type = self.type_register.get(model, self.is_input, UNSET)
        if not is_unset(field_type):
            return field_type

        db_field_type = type(self.field)
        raise TypeError(f"No type defined for field '{db_field_type.__name__}'"
                f" which has related model '{model._meta.object_name}'")


def get_field_type(field, type_register, is_input):
    db_field_type = type(field)

    if type_register:
        field_type = type_register.get(field.name, is_input, UNSET)
        if not is_unset(field_type):
            return field_type

        field_type = type_register.get(db_field_type, is_input, UNSET)
        if not is_unset(field_type):
            return field_type

        if field.is_relation:
            model = field.related_model

            field_type = type_register.get(model, is_input, UNSET)
            if not is_unset(field_type):
                return field_type

            # TODO: show field name
            raise TypeError(f"No type defined for field '{db_field_type.__name__}'"
                    f" which has related model '{model._meta.object_name}'")

    field_type = field_type_map.get(db_field_type, UNSET)
    if not is_unset(field_type):
        return field_type

    if field.is_relation:
        model = field.related_model
        raise TypeError(f"No type defined for Django model '{model._meta.object_name}'")

    # TODO: show field name
    raise TypeError(f"No type defined for '{db_field_type.__name__}'")


def is_in(item, item_list, default=False):
    if not item_list:
        return default
    return item in item_list


def process_fields(fields, model):
    field_names = []

    if fields is None:
        return field_names

    model_field_names = [field.name for field in model._meta.get_fields()]

    for field in fields:
        if isinstance(field, str):
            field_name = field
        else:
            raise TypeError('Type of field parameter should be str')
        if field_name not in model_field_names:
            raise AttributeError(f"Django model '{model._meta.object_name}' has no field '{field_name}'")
        field_names.append(field_name)
    return field_names


def get_model_fields(cls, model, fields, types, is_input, partial):
    if fields == []:
        return []

    field_names = process_fields(fields,  model)
    type_register = types

    model_fields = []
    for field in model._meta.get_fields():
        if not is_in(field.name, field_names, default=True):
            continue

        if is_input:
            if not field.editable:
                continue

            if field.is_relation:
                if field.many_to_many or field.one_to_many:
                    field_type = Optional[List[strawberry.ID]]
                    field_value = UNSET
                    model_fields.extend([
                        (f'{field.name}_add', field_type, field_value),
                        (f'{field.name}_set', field_type, field_value),
                        (f'{field.name}_remove', field_type, field_value),
                    ])
                else:
                    field_name = field.attname
                    field_type = strawberry.ID
                    if is_optional(field, is_input, partial):
                        field_type = Optional[field_type]
                    model_fields.append((field_name, field_type, UNSET))
                continue

        try:
            field_type = get_field_type(field, type_register, is_input)
        except TypeError:
            if not field.is_relation or type_register is None:
                raise
            field_type = LazyModelType(field, type_register, is_input)

        if field.is_relation:
            if field.many_to_many or field.one_to_many:
                field_type = List[field_type]
            field_value = strawberry_django_field(field_name=field.name)
        else:
            field_value = UNSET

        if is_optional(field, is_input, partial):
            field_type = Optional[field_type]

        model_fields.append((field.name, field_type, field_value))
    return model_fields


# post processing fields before passing them to strawberry
def update_fields(cls, model):
    for field_name, field in cls.__dict__.items():
        if not isinstance(field, DjangoField):
            continue

        django_field_name = field.field_name or field_name
        django_field = model._meta.get_field(django_field_name)
        is_m2m = django_field.many_to_many or django_field.one_to_many

        field = field.resolve(django_field.is_relation, is_m2m)
        setattr(cls, field_name, field)



import strawberry
from typing import Callable, List, Optional, Dict
import dataclasses
from . import utils, queries

@dataclasses.dataclass
class DjangoField:
    resolver: Callable
    field_name: Optional[str]
    kwargs: dict

    def resolve(self, is_relation, is_m2m):
        resolver = queries.resolvers.get_resolver(self.resolver, self.field_name, is_relation, is_m2m)
        field = strawberry.field(resolver, **self.kwargs)
        return field


def field(resolver=None, field_name=None, **kwargs):
    if resolver:
        resolver = queries.resolvers.get_resolver(resolver)
        return strawberry.field(resolver)

    return DjangoField(resolver, field_name, kwargs)

mutation = field



def add_hook(field, hook_name, hooks):
    if hooks is None:
        hooks = []
    if not isinstance(hooks, (list, tuple)):
        hooks = [hooks]
    field._hooks[hook_name] = hooks
    def register_hook(func):
        field._hooks[hook_name].append(func)
    setattr(field, hook_name, register_hook)

def add(**kwargs):
    def wrapper(field):
        field._hooks = {}
        for hook_name, hooks in kwargs.items():
            add_hook(field, hook_name, hooks)
        def call_hooks(hook_name, caller):
            for hook in field._hooks[hook_name]:
                caller(hook)
        field._call_hooks = call_hooks
        return field
    return wrapper

